@@ -4,7 +4,7 @@ package Classifier::MailParse;
 #
 # MailParse.pm --- Parse a mail message or messages into words
 #
-# Copyright (c) 2001-2003 John Graham-Cumming
+# Copyright (c) 2001-2004 John Graham-Cumming
 #
 #   This file is part of POPFile
 #
@@ -43,6 +43,10 @@ my $eksc = "(?:$ksc5601|[\x81-\xC6][\x41-\xFE])"; #extended ksc
 
 # These are used for Japanese support
 
+my %encoding_candidates = (
+    'Nihongo' => [ 'shiftjis', 'euc-jp', '7bit-jis' ]
+);
+
 my $ascii = '[\x00-\x7F]'; # ASCII chars
 my $two_bytes_euc_jp = '(?:[\x8E\xA1-\xFE][\xA1-\xFE])'; # 2bytes EUC-JP chars
 my $three_bytes_euc_jp = '(?:\x8F[\xA1-\xFE][\xA1-\xFE])'; # 3bytes EUC-JP chars
@@ -155,15 +159,8 @@ sub new
 
     $self->{quickmagnets__}      = {};
 
-    # These store the current HTML background color and font color to
-    # detect "invisible ink" used by spammers
-
-    $self->{htmlbackcolor__} = map_color( $self, 'white' );
-    $self->{htmlbodycolor__} = map_color( $self, 'white' );
-    $self->{htmlfontcolor__} = map_color( $self, 'black' );
-
-    # store the tag that set the foreground/background color so the color can be
-    # unset when the tag closes
+    # store the tag that set the foreground/background color so the
+    # color can be unset when the tag closes
 
     $self->{cssfontcolortag__} = '';
     $self->{cssbackcolortag__} = '';
@@ -173,8 +170,9 @@ sub new
 
     $self->{htmlcolordistance__} = 0;
 
-    # This is a mapping between HTML color names and HTML hexadecimal color values used by the
-    # map_color value to get canonical color values
+    # This is a mapping between HTML color names and HTML hexadecimal
+    # color values used by the map_color value to get canonical color
+    # values
 
     $self->{color_map__} = { 'aliceblue','f0f8ff', 'antiquewhite','faebd7', 'aqua','00ffff', 'aquamarine','7fffd4', 'azure','f0ffff', # PROFILE BLOCK START
         'beige','f5f5dc', 'bisque','ffe4c4', 'black','000000', 'blanchedalmond','ffebcd', 'blue','0000ff', 'blueviolet','8a2be2',
@@ -202,6 +200,15 @@ sub new
         'tan','d2b48c', 'teal','008080', 'thistle','d8bfd8', 'tomato','ff6347', 'turquoise','40e0d0', 'violet','ee82ee', 'wheat','f5deb3',
         'white','ffffff', 'whitesmoke','f5f5f5', 'yellow','ffff00', 'yellowgreen','9acd32' }; # PROFILE BLOCK STOP
 
+    # These store the current HTML background color and font color to
+    # detect "invisible ink" used by spammers
+
+    my $result = bless $self, $type;
+
+    $self->{htmlbackcolor__} = $self->map_color( 'white' );
+    $self->{htmlbodycolor__} = $self->map_color( 'white' );
+    $self->{htmlfontcolor__} = $self->map_color( 'black' );
+
     $self->{content_type__} = '';
     $self->{base64__}       = '';
     $self->{in_html_tag__}  = 0;
@@ -210,11 +217,14 @@ sub new
     $self->{in_headers__}   = 0;
 
     # This is used for switching on/off language specific functionality
-    $self->{lang__} = '';
 
+    $self->{lang__} = '';
     $self->{first20__}      = '';
 
-    return bless $self, $type;
+    # For support Quoted Printable in Japanese text, save encoded text in multiple lines
+    $self->{prev__} = '';
+
+    return $result;
 }
 
 # ---------------------------------------------------------------------------------------------
@@ -239,21 +249,21 @@ sub get_color__
             if ( $word eq $self->{color_idmap__}{$i} ) {
                 $id = $i;
                 last;
-	    }
-	}
+            }
+        }
 
         if ( defined( $id ) ) {
             my @buckets = $self->{bayes__}->get_buckets( $self->{color__} );
 
-	    return $self->{bayes__}->get_bucket_color( $self->{color__},
+            return $self->{bayes__}->get_bucket_color( $self->{color__},
                 $self->{bayes__}->get_top_bucket__(
                     $self->{color_userid__},
                     $id,
                     $self->{color_matrix__},
-		    \@buckets ) );
-	} else {
+                    \@buckets ) );
+        } else {
             return 'black';
-	}
+        }
     }
 }
 
@@ -328,17 +338,81 @@ sub map_color
 {
     my ( $self, $color ) = @_;
 
-    # The canonical form is lowercase hexadecimal, so start by lowercasing and stripping any
-    # initial #
+    # The canonical form is lowercase hexadecimal, so start by
+    # lowercasing and stripping any initial #
 
     $color = lc( $color );
-    $color =~ s/^#//;
 
     # Map color names to hexadecimal values
 
     if ( defined( $self->{color_map__}{$color} ) ) {
         return $self->{color_map__}{$color};
     } else {
+
+        # Do this after checking the color map, as there is no "#blue" color
+        # TODO: The #, however, is optional in IE.. Do we pseudo-word this?
+
+        $color =~ s/^#//;
+
+        my $old_color = $color;
+
+        # Due to a bug/feature in Microsoft Internet Explorer it's
+        # possible to use invalid hexadecimal colors where the number
+        # 0 is replaced by any other character and if the hex has an
+        # uneven multiple of 3 number of characters it is padded on
+        # the right with 0s and if the hex is too long, it is divided
+        # into even triplets with the leftmost characters in the
+        # triplets being significant. Short (1-char) triplets are
+        # left-padded with 0's
+
+        my $quotient = ((length($color) - (length($color) % 3)) / 3 );
+
+        # We go one higher than the quotient if the length isn't an
+        # even multiple of 3
+
+        $quotient += 1 if (length($color) % 3);
+
+        # right-pad entire value to get past the next
+        # full multiple of the quotient ("abc abc abc"
+        # needs at least one more character to make
+        # three even triplets)
+
+        $color .= "00" . "0" x $quotient;
+
+        # even length RGB triplets
+        my ($r, $g, $b) = ($color =~ /(.{$quotient})(.{$quotient})(.{$quotient})/);
+
+        print "$r $g $b\n" if $self->{debug_};
+
+        # left-trim very long triplets to 4 bytes
+        $r =~ s/.*(.{8})$/$1/;
+        $g =~ s/.*(.{8})$/$1/;
+        $b =~ s/.*(.{8})$/$1/;
+
+        # right-trim long triplets to get the first two bytes
+        $r =~ s/(..).*/$1/;
+        $g =~ s/(..).*/$1/;
+        $b =~ s/(..).*/$1/;
+
+        # left-pad short triplets (eg FFF -> 0F0F0F)
+        $r = '0' . $r if (length($r) == 1);
+        $g = '0' . $g if (length($g) == 1);
+        $b = '0' . $b if (length($b) == 1);
+
+        $color = "$r$g$b"; # here is our real color value
+
+        # Any non-hex values remaining get 0'd out
+        $color =~ s/[^0-9a-f]/0/g;
+
+        print "hex color $color\n" if ($self->{debug__});
+        print "flex-hex detected\n" if ($self->{debug__} && ($color ne $old_color));
+
+        # Add pseudo-word anytime flex hex detected
+
+        if ( $color ne $old_color ) {
+            $self->update_pseudoword( 'trick:flexhex', $old_color, 0, '' );
+        }
+
         return $color;
     }
 }
@@ -392,7 +466,7 @@ sub update_pseudoword
                 my $color = $self->get_color__($mword);
                 my $to    = "<b><font color=\"$color\"><a title=\"$mword\">$literal</a></font></b>";
                 $self->{ut__} .= $to . ' ';
-	    }
+            }
         }
 
         $self->increment_word( $mword );
@@ -501,9 +575,9 @@ sub add_line
                     # HTML entities confilict with DBCS chars. Replace entities with blanks.
 
                     if ( $self->{lang__} eq 'Korean' ) {
-                    	$to = ' ';
+                            $to = ' ';
                     } else {
-                	$to = chr($to);
+                        $to = chr($to);
                     }
                     $line       =~ s/$from/$to/g;
                     $self->{ut__} =~ s/$from/$to/g;
@@ -606,7 +680,7 @@ sub add_line
                         }
 
                         update_word($self,$1, $encoded, '', '[_\-,\.\"\'\)\?!:;\/ &\t\n\r]', $prefix) if (length $1 >= 2);
-		    }
+                    }
                 } else {
 
                     # Only care about words between 3 and 45 characters since short words like
@@ -675,7 +749,7 @@ sub update_tag
             $self->compute_html_color_distance();
         }
 
-        if ( $tag =~ /^$self->{cssbackcolortag__}$/i ) {
+        if ( lc( $tag ) eq $self->{cssbackcolortag__} ) {
             $self->{htmlbackcolor__} = $self->{htmlbodycolor__};
             $self->{cssbackcolortag__} = '';
 
@@ -684,7 +758,7 @@ sub update_tag
             print "CSS back color reset to $self->{htmlbackcolor__} (tag closed: $tag)\n" if ( $self->{debug__} );
         }
 
-        if ( $tag =~ /^$self->{cssfontcolortag__}$/i ) {
+        if ( lc( $tag ) eq $self->{cssfontcolortag__} ) {
             $self->{htmlfontcolor__} = map_color( $self, 'black' );
             $self->{cssfontcolortag__} = '';
 
@@ -749,9 +823,15 @@ sub update_tag
             # Adding strings from these, even if they appear to be hostnames, may or
             # may not be beneficial
 
-            if ($value =~ /^cid\:/i )
-            {
-                # TODO: Decide what to do here, ignoring CID's for now
+            if ($value =~ /^(cid)\:/i ) {
+
+                # Add a pseudo-word when CID source links are detected
+
+                $self->update_pseudoword( 'html', 'cidsrc' );
+
+                # TODO: I've seen virus messages try to use a CID: href
+
+
             } else {
 
                 my $host = add_url( $self, $value, $encoded, $quote, $end_quote, '' );
@@ -1334,7 +1414,6 @@ sub parse_html
 # HTML version of message if color__ is set
 #
 # $file     The file to open and parse
-# $lang     Pass in the current interface language for language specific parsing
 # $max_size The maximum size of message to parse, or 0 for unlimited
 # $reset    If set to 0 then the list of words from a previous parse is not reset, this
 #           can be used to do multiple parses and build a single word list.  By default
@@ -1343,16 +1422,11 @@ sub parse_html
 # ---------------------------------------------------------------------------------------------
 sub parse_file
 {
-    # $lang is used for switching on/off language specific functionality
-
-    my ( $self, $file, $lang, $max_size, $reset ) = @_;
+    my ( $self, $file, $max_size, $reset ) = @_;
 
     $reset    = 1 if ( !defined( $reset    ) );
     $max_size = 0 if ( !defined( $max_size ) );
 
-    $lang = '' unless ( defined($lang) );
-
-    $self->{lang__} = $lang;
     $self->start_parse( $reset );
 
     my $size_read = 0;
@@ -1369,7 +1443,7 @@ sub parse_file
         if ( ( $max_size > 0 ) &&
              ( $size_read > $max_size ) ) {
             last;
-	}
+        }
     }
 
     close MSG;
@@ -1526,6 +1600,32 @@ sub parse_line
 
             print ">>> $line" if $self->{debug__};
 
+            # Decode quoted-printable
+            if ( !$self->{in_headers__} && $self->{encoding__} =~ /quoted\-printable/i) {
+                if ( $self->{lang__} eq 'Nihongo') {
+                    if ( $line =~ /=\r\n$/ ) {
+                        # Encoded in multiple lines
+                        $line =~ s/=\r\n$//g;
+                        $self->{prev__} .= $line;
+                        next;
+                    } else {
+                        $line = $self->{prev__} . $line;
+                        $self->{prev__} = '';
+                    }
+                }
+                $line = decode_qp( $line );
+            }
+
+            # Decode \x??
+            if ( $self->{lang__} eq 'Nihongo' && !$self->{in_headers__} ) {
+                $line =~ s/\\x([8-9A-F][A-F0-9])/pack("C", hex($1))/eig;
+            }
+
+            if ( $self->{lang__} eq 'Nihongo' ) {
+                $line = convert_encoding( $line, $self->{charset__}, 'euc-jp', '7bit-jis', @{$encoding_candidates{$self->{lang__}}} );
+                $line = parse_line_with_kakasi( $self, $line );
+            }
+
             if ($self->{color__} ne '' ) {
 
                 if (!$self->{in_html_tag__}) {
@@ -1654,14 +1754,6 @@ sub parse_line
 
             next if ( !defined($line) );
 
-            # Decode quoted-printable
-
-            if ( $self->{encoding__} =~ /quoted\-printable/i ) {
-                $line       = decode_qp( $line );
-                $line       =~ s/[\r\n]+$/ /g;
-                $self->{ut__} = decode_qp( $self->{ut__} ) if ( $self->{color__} ne '' );
-            }
-
             parse_html( $self, $line, 0 );
         }
     }
@@ -1711,7 +1803,8 @@ sub clear_out_base64
 
 # ---------------------------------------------------------------------------------------------
 #
-# decode_string - Decode MIME encoded strings used in the header lines in email messages
+# decode_string - Decode MIME encoded strings used in the header lines
+# in email messages
 #
 # $mystring     - The string that neeeds decode
 #
@@ -1719,27 +1812,34 @@ sub clear_out_base64
 #
 # =?charset?[BQ]?text?=
 #
-# $lang     Pass in the current interface language for language specific encoding conversion
-# A B indicates base64 encoding, a Q indicates quoted printable encoding
+# $lang Pass in the current interface language for language specific
+# encoding conversion A B indicates base64 encoding, a Q indicates
+# quoted printable encoding
 # ---------------------------------------------------------------------------------------------
 sub decode_string
 {
-    # I choose not to use "$mystring = MIME::Base64::decode( $1 );" because some spam mails
-    # have subjects like: "Subject: adjpwpekm =?ISO-8859-1?Q?=B2=E1=A4=D1=AB=C7?= dopdalnfjpw".
-    # Therefore, it will be better to store the decoded text in a temporary variable and substitute
-    # the original string with it later. Thus, this subroutine returns the real decoded result.
+    # I choose not to use "$mystring = MIME::Base64::decode( $1 );"
+    # because some spam mails have subjects like: "Subject: adjpwpekm
+    # =?ISO-8859-1?Q?=B2=E1=A4=D1=AB=C7?= dopdalnfjpw".  Therefore, it
+    # will be better to store the decoded text in a temporary variable
+    # and substitute the original string with it later. Thus, this
+    # subroutine returns the real decoded result.
 
     my ( $self, $mystring, $lang ) = @_;
 
     my $decode_it = '';
+    my $charset = '';
+
+    $lang = $self->{lang__} if ( !defined( $lang ) || ( $lang eq '' ) );
 
     while ( $mystring =~ /=\?([\w-]+)\?(B|Q)\?(.*?)\?=/ig ) {
         if ($2 eq "B" || $2 eq "b") {
+            $charset = $1;
             $decode_it = decode_base64( $3 );
 
             # for Japanese header
-            if ((uc($1) eq "ISO-2022-JP") && ( $lang eq 'Nihongo' )) {
-                $decode_it = convert_encoding($decode_it, "iso-2022-jp", "euc-jp");
+            if ($lang eq 'Nihongo') {
+                $decode_it = convert_encoding( $decode_it, $charset, 'euc-jp', '7bit-jis', @{$encoding_candidates{$self->{lang__}}} );
             }
 
             $mystring =~ s/=\?[\w-]+\?B\?(.*?)\?=/$decode_it/i;
@@ -1750,8 +1850,8 @@ sub decode_string
                 $decode_it = decode_qp( $decode_it );
 
                 # for Japanese header
-                if ((uc($1) eq "ISO-2022-JP") && ( $lang eq 'Nihongo' )) {
-                    $decode_it = convert_encoding($decode_it, "iso-2022-jp", "euc-jp");
+                if ($lang eq 'Nihongo') {
+                    $decode_it = convert_encoding( $decode_it, $charset, 'euc-jp', '7bit-jis', @{$encoding_candidates{$self->{lang__}}} );
                 }
 
                 $mystring =~ s/=\?[\w-]+\?Q\?(.*?)\?=/$decode_it/i;
@@ -1845,7 +1945,7 @@ sub parse_header
             if ( $self->{from__} eq '' ) {
                 $self->{from__} = $argument;
                 $self->{from__} =~ s/[\t\r\n]//g;
-	    }
+            }
         }
 
         if ( $header =~ /^To$/i ) {
@@ -1853,7 +1953,7 @@ sub parse_header
             if ( $self->{to__} eq '' ) {
                 $self->{to__} = $argument;
                 $self->{to__} =~ s/[\t\r\n]//g;
-	    }
+            }
         }
 
         if ( $header =~ /^Cc$/i ) {
@@ -1861,7 +1961,7 @@ sub parse_header
             if ( $self->{cc__} eq '' ) {
                 $self->{cc__} = $argument;
                 $self->{cc__} =~ s/[\t\r\n]//g;
-	    }
+            }
         }
 
         while ( $argument =~ s/<([[:alpha:]0-9\-_\.]+?@([[:alpha:]0-9\-_\.]+?))>// )  {
@@ -1883,12 +1983,18 @@ sub parse_header
         $prefix = 'subject';
         $argument = $self->decode_string( $argument, $self->{lang__} );
         if ( $self->{subject__} eq '' ) {
+
+            # In Japanese mode, parse subject with kakasi
+
+            $argument = parse_line_with_kakasi( $self, $argument ) if ( $self->{lang__} eq 'Nihongo' && $argument ne '' );
+
             $self->{subject__} = $argument;
             $self->{subject__} =~ s/[\t\r\n]//g;
-	}
+        }
     }
 
     $self->{date__} = $argument if ( $header =~ /^Date$/i );
+
     if ( $header =~ /^X-Spam-Status$/i) {
 
         # We have found a header added by SpamAssassin. We expect to
@@ -1920,6 +2026,7 @@ sub parse_header
 
     if ( $header =~ /^Content-Type$/i ) {
         if ( $argument =~ /charset=\"?([^\"\r\n\t ]{1,40})\"?/ ) {
+            $self->{charset__} = $1;
             update_word( $self, $1, 0, '' , '', 'charset' );
         }
 
@@ -1979,7 +2086,7 @@ sub parse_header
 
     if ( $header =~ /^Content-Disposition$/i ) {
         $self->handle_disposition( $argument );
-	return ( $mime, $encoding );
+        return ( $mime, $encoding );
     }
 
     add_line( $self, $argument, 0, $prefix );
@@ -2301,15 +2408,76 @@ sub mangle
 # $string       The string to be converted
 # $from         Original encoding
 # $to           The encoding which the string is converted to
+# $default      The default encoding that is used when $from is invalid or not defined
+# @candidates   Candidate encodings for guessing
 # ---------------------------------------------------------------------------------------------
 sub convert_encoding
 {
-    my ( $string, $from, $to ) = @_;
+    my ( $string, $from, $to, $default, @candidates ) = @_;
+
     require Encode;
+    require Encode::Guess;
 
-    Encode::from_to($string, $from, $to);
+    # First, guess the encoding.
 
+    my $enc = Encode::Guess::guess_encoding( $string, @candidates );
+
+    if(ref $enc){
+       $from= $enc->name;
+    } else {
+
+        # If guess does not work, check whether $from is valid.
+
+        if (!(Encode::resolve_alias($from))) {
+
+            # Use $default as $from when $from is invalid.
+
+            $from = $default;
+        }
+    }
+
+    unless ($from eq $to) {
+        my ($orig_string) = $string;
+
+        # Workaround for Encode::Unicode error bug.
+        eval {
+            Encode::from_to($string, $from, $to);
+        };
+        $string = $orig_string if ($@);
+    }
     return $string;
 }
 
+# ---------------------------------------------------------------------------------------------
+#
+# parse_line_with_kakasi
+#
+# Parse a line with Kakasi
+#
+# Japanese needs to be parsed by language processing filter, "Kakasi"
+# before it is passed to Bayes classifier because words are not splitted
+# by spaces.
+#
+# $line          The line to be parsed
+#
+# ---------------------------------------------------------------------------------------------
+sub parse_line_with_kakasi
+{
+    my ( $self, $line ) = @_;
+
+    # This is used to parse Japanese
+    require Text::Kakasi;
+
+    # Split Japanese line into words using Kakasi Wakachigaki
+    # mode(-w is passed to Kakasi as argument). Both input and ouput
+    # encoding are EUC-JP.
+
+    Text::Kakasi::getopt_argv("kakasi", "-w -ieuc -oeuc");
+    $line = Text::Kakasi::do_kakasi($line);
+    Text::Kakasi::close_kanwadict();
+
+    return $line;
+}
+
+
 1;
