@@ -8,7 +8,7 @@ use POPFile::Module;
 #
 # WordMangle.pm --- Mangle words for better classification
 #
-# Copyright (c) 2001-2003 John Graham-Cumming
+# Copyright (c) 2001-2004 John Graham-Cumming
 #
 #   This file is part of POPFile
 #
@@ -86,7 +86,7 @@ sub load_stopwords
 
         close STOPS;
     } else { 
-        $self->log_( "Failed to open stopwords file" );
+        $self->log_( 0, "Failed to open stopwords file" );
     }
 }
 
@@ -179,7 +179,7 @@ sub add_stopword
             return 0;
         }
     } else {
-        if ( ( $stopword !~ /:/ ) && ( $stopword =~ /[^[:lower:]\-_\.\@0-9]/i ) ) {
+        if ( ( $stopword !~ /:/ ) && ( $stopword =~ /[^[:alpha:]\-_\.\@0-9]/i ) ) {
             return 0;
         }
     }
@@ -207,7 +207,7 @@ sub remove_stopword
             return 0;
         }
     } else {
-        if ( ( $stopword !~ /:/ ) && ( $stopword =~ /[^[:lower:]\-_\.\@0-9]/i ) ) {
+        if ( ( $stopword !~ /:/ ) && ( $stopword =~ /[^[:alpha:]\-_\.\@0-9]/i ) ) {
             return 0;
         }
     }
