@@ -11,7 +11,7 @@ use POPFile::Module;
 # register specific parameters with this module.  This module also handles
 # POPFile's command line parsing
 #
-# Copyright (c) 2001-2003 John Graham-Cumming
+# Copyright (c) 2001-2004 John Graham-Cumming
 #
 #   This file is part of POPFile
 #
@@ -47,11 +47,11 @@ sub new
     my $type = shift;
     my $self = POPFile::Module->new();
 
-    # All the current configuration parameters are stored in this hash which
-    # is intended to be globally accessed by modules that make use of this module,
-    # to register a configuration default entries are made in this hash in the form
+    # This hash has indexed by parameter name and has two fields:
     #
-    # $self->{configuration_parameters__}{parameter}
+    # value         The current value
+    # default       The default value
+
     $self->{configuration_parameters__} = {};
 
     # Name of the PID file that we created
@@ -70,6 +70,12 @@ sub new
 
     $self->{save_needed__} = 0;
 
+    # We track when out start() is called so that we know when the modules
+    # are done setting the default values so that we know which have default
+    # and which do not
+
+    $self->{started__} = 0;
+
     # Local copies of POPFILE_ROOT and POPFILE_USER
 
     $self->{popfile_root__} = $ENV{POPFILE_ROOT} || './';
@@ -98,21 +104,6 @@ sub initialize
 
     $self->config_( 'piddir', './' );
 
-    # This counter is used when creating unique IDs for message stored
-    # in the history.  The history message files have the format
-    #
-    # popfile{download_count}={message_count}.msg
-    #
-    # Where the download_count is derived from this value and the
-    # message_count is a local counter within that download, for sorting
-    # purposes must sort on download_count and then message_count
-    #
-    # download_count is incremented every time POPFile forks to
-    # start a session for downloading messages (see Proxy::Proxy::service
-    # for details)
-
-    $self->global_config_( 'download_count', 0 );
-
     # The default timeout in seconds for POP3 commands
 
     $self->global_config_( 'timeout', 60 );
@@ -126,6 +117,18 @@ sub initialize
 
     $self->global_config_( 'message_cutoff', 100000 );
 
+    # Register for the TICKD message which is sent hourly by the
+    # Logger module.   We use this to hourly save the configuration file
+    # so that POPFile's configuration is saved in case of a hard crash.
+    #
+    # This is particularly needed by the IMAP module which stores some
+    # state related information in the configuration parameters.  Note that
+    # because of the save_needed__ bool there wont be any write to the
+    # disk unless a configuration parameter has been changed since the
+    # last save.  (see parameter())
+
+    $self->mq_register_( 'TICKD', $self );
+
     return 1;
 }
 
@@ -140,10 +143,14 @@ sub start
 {
     my ( $self ) = @_;
 
-    # Check to see if the PID file is present, if it is then another POPFile
-    # may be running, warn the user and terminate
+    $self->{started__} = 1;
 
-    $self->{pid_file__} = $self->get_user_path( $self->config_( 'piddir' ) . 'popfile.pid' );
+    # Check to see if the PID file is present, if it is then another POPFile
+    # may be running, warn the user and terminate, note the 0 at the end
+    # means that we allow the piddir to be absolute and outside the user
+    # sandbox
+
+    $self->{pid_file__} = $self->get_user_path( $self->config_( 'piddir' ) . 'popfile.pid', 0 );
 
     if (defined($self->live_check_())) {
         return 0;
@@ -176,7 +183,7 @@ sub service
 
         if ( !$self->check_pid_() ) {
             $self->write_pid_();
-            $self->log_("New POPFile instance detected and signalled")
+            $self->log_( 0, "New POPFile instance detected and signalled" );
         }
     }
 
@@ -199,6 +206,22 @@ sub stop
     $self->delete_pid_();
 }
 
+# ---------------------------------------------------------------------------------------------
+#
+# deliver
+#
+# Called by the message queue to deliver a message
+#
+# ---------------------------------------------------------------------------------------------
+sub deliver
+{
+    my ( $self, $type, @message ) = @_;
+
+    if ( $type eq 'TICKD' ) {
+        $self->save_configuration();
+    }
+}
+
 # ---------------------------------------------------------------------------------------------
 #
 # live_check_
@@ -326,7 +349,7 @@ sub parse_command_line
     #
     # So its possible to do
     #
-    # --set bayes_param=value --set=-bayes_parem=value --set -bayes_parem=value -- -bayes_parem value
+    # --set bayes_param=value --set=-bayes_param=value --set -bayes_param=value -- -bayes_param value
 
     if ( !GetOptions( "set=s" => \@set_options ) ) {
         return 0;
@@ -398,11 +421,13 @@ sub upgrade_parameter__
 {
     my ( $self, $parameter ) = @_;
 
-    # This table maps from the old parameter to the new one, for example the old
-    # xpl parameter which controls insertion of the X-POPFile-Link header in email
-    # is now called GLOBAL_xpl and is accessed through POPFile::Module::global_config_
-    # The old piddir parameter is now config_piddir and is accessed through either config_
-    # if accessed from the config module or through module_config_ from outside
+    # This table maps from the old parameter to the new one, for
+    # example the old xpl parameter which controls insertion of the
+    # X-POPFile-Link header in email is now called GLOBAL_xpl and is
+    # accessed through POPFile::Module::global_config_ The old piddir
+    # parameter is now config_piddir and is accessed through either
+    # config_ if accessed from the config module or through
+    # module_config_ from outside
 
     my %upgrades = ( # PROFILE BLOCK START
 
@@ -420,7 +445,6 @@ sub upgrade_parameter__
                      'debug',                    'GLOBAL_debug',
                      'msgdir',                   'GLOBAL_msgdir',
                      'timeout',                  'GLOBAL_timeout',
-                     'download_count',           'GLOBAL_download_count',
 
                      # Parameters that are now handled by POPFile::Logger
 
@@ -437,10 +461,6 @@ sub upgrade_parameter__
 
                      # Parameters that are now handled by UI::HTML
 
-                     'archive',                  'html_archive',
-                     'archive_classes',          'html_archive_classes',
-                     'archive_dir',              'html_archive_dir',
-                     'history_days',             'html_history_days',
                      'language',                 'html_language',
                      'last_reset',               'html_last_reset',
                      'last_update_check',        'html_last_update_check',
@@ -452,6 +472,19 @@ sub upgrade_parameter__
                      'test_language',            'html_test_language',
                      'update_check',             'html_update_check',
                      'ui_port',                  'html_port',
+
+                     # Parameters the have moved from the UI::HTML to
+                     # POPFile::History
+
+                     'archive',                  'history_archive',
+                     'archive_classes',          'history_archive_classes',
+                     'archive_dir',              'history_archive_dir',
+                     'history_days',             'history_history_days',
+                     'html_archive',             'history_archive',
+                     'html_archive_classes',     'history_archive_classes',
+                     'html_archive_dir',         'history_archive_dir',
+                     'html_history_days',        'history_history_days',
+
     ); # PROFILE BLOCK STOP
 
     if ( defined( $upgrades{$parameter} ) ) {
@@ -473,6 +506,8 @@ sub load_configuration
 {
     my ( $self ) = @_;
 
+    $self->{started__} = 1;
+
     if ( open CONFIG, '<' . $self->get_user_path( 'popfile.cfg' ) ) {
         while ( <CONFIG> ) {
             s/(\015|\012)//g;
@@ -484,9 +519,8 @@ sub load_configuration
                 $parameter = $self->upgrade_parameter__($parameter);
 
                 if ( defined( $self->{configuration_parameters__}{$parameter} ) ) {
-                    $self->{configuration_parameters__}{$parameter} = $value;
-	        } else {
-                    $self->log_( "Discarded unknown parameter '$parameter' from popfile.cfg" );
+                    $self->{configuration_parameters__}{$parameter}{value} = $value;
+                } else {
                     $self->{deprecated_parameters__}{$parameter} = $value;
                 }
             }
@@ -517,7 +551,7 @@ sub save_configuration
         $self->{save_needed__} = 0;
 
         foreach my $key (sort keys %{$self->{configuration_parameters__}}) {
-            print CONFIG "$key $self->{configuration_parameters__}{$key}\n";
+            print CONFIG "$key $self->{configuration_parameters__}{$key}{value}\n";
         }
 
         close CONFIG;
@@ -531,20 +565,22 @@ sub save_configuration
 # Resolve a path relative to POPFILE_USER or POPFILE_ROOT
 #
 # $path              The path to resolve
+# $sandbox           Set to 1 if this path must be sandboxed (i.e. absolute
+#                    paths and paths containing .. are not accepted).
 #
 # ---------------------------------------------------------------------------------------------
 sub get_user_path
 {
-    my ( $self, $path ) = @_;
+    my ( $self, $path, $sandbox ) = @_;
 
-    return $self->path_join__( $self->{popfile_user__}, $path );
+    return $self->path_join__( $self->{popfile_user__}, $path, $sandbox );
 }
 
 sub get_root_path
 {
-    my ( $self, $path ) = @_;
+    my ( $self, $path, $sandbox ) = @_;
 
-    return $self->path_join__( $self->{popfile_root__}, $path );
+    return $self->path_join__( $self->{popfile_root__}, $path, $sandbox );
 }
 
 # ---------------------------------------------------------------------------------------------
@@ -553,16 +589,32 @@ sub get_root_path
 #
 # Join two paths togther
 #
-# $left                The LHS
-# $right               The RHS
+# $left              The LHS
+# $right             The RHS
+# $sandbox           Set to 1 if this path must be sandboxed (i.e. absolute
+#                    paths and paths containing .. are not accepted).
 #
 # ---------------------------------------------------------------------------------------------
 sub path_join__
 {
-    my ( $self, $left, $right ) = @_;
+    my ( $self, $left, $right, $sandbox ) = @_;
 
-    if ( ( $right =~ /^\// ) || ( $right =~ /^[A-Za-z]:[\/\\]/ ) ) {
-        return $right;
+    $sandbox = 1 if ( !defined( $sandbox ) );
+
+    if ( ( $right =~ /^\// ) ||
+         ( $right =~ /^[A-Za-z]:[\/\\]/ ) ||
+         ( $right =~ /\\\\/ ) ) {
+        if ( $sandbox ) {
+            $self->log_( 0, "Attempt to access path $right outside sandbox" );
+            return undef;
+        } else {
+            return $right;
+        }
+    }
+
+    if ( $sandbox && ( $right =~ /\.\./ ) ) {
+        $self->log_( 0, "Attempt to access path $right outside sandbox" );
+        return undef;
     }
 
     $left  =~ s/\/$//;
@@ -585,14 +637,36 @@ sub path_join__
 # ---------------------------------------------------------------------------------------------
 sub parameter
 {
-  my ( $self, $name, $value ) = @_;
+    my ( $self, $name, $value ) = @_;
 
-  if ( defined( $value ) ) {
-    $self->{save_needed__} = 1;
-    $self->{configuration_parameters__}{$name} = $value;
-  }
+    if ( defined( $value ) ) {
+        $self->{save_needed__} = 1;
+        $self->{configuration_parameters__}{$name}{value} = $value;
+        if ( $self->{started__} == 0 ) {
+            $self->{configuration_parameters__}{$name}{default} = $value;
+        }
+    }
 
-  return $self->{configuration_parameters__}{$name};
+    return $self->{configuration_parameters__}{$name}{value};
+}
+
+# ---------------------------------------------------------------------------------------------
+#
+# is_default
+#
+# Returns whether the parameter has the default value or not
+#
+# $name          Name of the parameter
+#
+# Returns 1 if the parameter still has its default value
+#
+# ---------------------------------------------------------------------------------------------
+sub is_default
+{
+    my ( $self, $name ) = @_;
+
+    return ( $self->{configuration_parameters__}{$name}{value} eq
+             $self->{configuration_parameters__}{$name}{default} );
 }
 
 # GETTERS
