@@ -4,7 +4,7 @@ package Proxy::Proxy;
 #
 # This module implements the base class for all POPFile proxy Modules
 #
-# Copyright (c) 2001-2003 John Graham-Cumming
+# Copyright (c) 2001-2004 John Graham-Cumming
 #
 #   This file is part of POPFile
 #
@@ -36,8 +36,6 @@ use IO::Select;
 # A handy variable containing the value of an EOL for networks
 my $eol = "\015\012";
 
-use POSIX ":sys_wait_h";
-
 #----------------------------------------------------------------------------
 # new
 #
@@ -51,25 +49,19 @@ sub new
     my $type = shift;
     my $self = POPFile::Module->new();
 
-    # A reference to the classifier
+    # A reference to the classifier and history
 
     $self->{classifier__}     = 0;
-
-    # List of file handles to read from active children, this
-    # maps the PID for each child to its associated pipe handle
-
-    $self->{children__}        = {};
+    $self->{history__}        = 0;
 
     # Reference to a child() method called to handle a proxy
-    # connection, reference to flush_child_data() method used
-    # to clear out pipes
+    # connection
 
     $self->{child_}            = 0;
-    $self->{flush_child_data_} = \&flush_child_data_;
 
     # Holding variable for MSWin32 pipe handling
 
-    $self->{pipe_cache__};
+    $self->{pipe_cache__} = {};
 
     # This is where we keep the session with the Classifier::Bayes
     # module
@@ -92,6 +84,9 @@ sub new
     #
     # $self->{good_response_}            = '';
 
+    # Connect Banner returned by the real server
+    $self->{connect_banner__} = '';
+
     return bless $self, $type;
 }
 
@@ -109,6 +104,12 @@ sub initialize
 
     $self->config_( 'enabled', 1 );
 
+    # The following parameters are for SOCKS proxy handling on outbound
+    # connections
+
+    $self->config_( 'socks_server', '' );
+    $self->config_( 'socks_port',   1080 );
+
     return 1;
 }
 
@@ -129,14 +130,15 @@ sub start
     # Open the socket used to receive request for proxy service
 
     $self->{server__} = IO::Socket::INET->new( Proto     => 'tcp', # PROFILE BLOCK START
-                                    $self->config_( 'local' ) == 1 ? (LocalAddr => 'localhost') : (),
+                                    ($self->config_( 'local' ) || 0) == 1 ? (LocalAddr => 'localhost') : (),
                                     LocalPort => $self->config_( 'port' ),
                                     Listen    => SOMAXCONN,
                                     Reuse     => 1 ); # PROFILE BLOCK STOP
 
+    my $name = $self->name();
+
     if ( !defined( $self->{server__} ) ) {
         my $port = $self->config_( 'port' );
-        my $name = $self->name();
         print STDERR <<EOM; # PROFILE BLOCK START
 
 \nCouldn't start the $name proxy because POPFile could not bind to the
@@ -155,6 +157,13 @@ EOM
 
     $self->{selector__} = new IO::Select( $self->{server__} );
 
+    # Tell the UI about the SOCKS parameters
+
+    $self->register_configuration_item_( 'configuration',  # PROFILE BLOCK START
+                                         $name . '_socks_configuration',
+                                         'socks-widget.thtml',
+                                         $self );          # PROFILE BLOCK STOP
+
     return 1;
 }
 
@@ -163,7 +172,7 @@ EOM
 # stop
 #
 # Called when POPFile is closing down, this is the last method that will get called before
-# the object is destroyed.  There is not return value from stop().
+# the object is destroyed.  There is no return value from stop().
 #
 # ---------------------------------------------------------------------------------------------
 sub stop
@@ -178,131 +187,6 @@ sub stop
     # and all the reading ends of pipes to active children
 
     close $self->{server__} if ( defined( $self->{server__} ) );
-
-    for my $kid (keys %{$self->{children__}}) {
-        close $self->{children__}{$kid};
-        delete $self->{children__}{$kid};
-    }
-}
-
-# ---------------------------------------------------------------------------------------------
-#
-# reaper
-#
-# Called when a child process terminates somewhere in POPFile.  The object should check
-# to see if it was one of its children and do any necessary processing by calling waitpid()
-# on any child handles it has
-#
-# There is no return value from this method
-#
-# ---------------------------------------------------------------------------------------------
-sub reaper
-{
-    my ( $self ) = @_;
-
-    # Look for children that have completed and then flush the data from their
-    # associated pipe and see if any of our children have data ready to read from their pipes,
-
-    my @kids = keys %{$self->{children__}};
-
-    if ( $#kids >= 0 ) {
-        for my $kid (@kids) {
-            if ( waitpid( $kid, &WNOHANG ) == $kid ) {
-                $self->{flush_child_data_}( $self, $self->{children__}{$kid} );
-                close $self->{children__}{$kid};
-                delete $self->{children__}{$kid};
-
-                $self->log_( "Done with $kid (" . scalar(keys %{$self->{children__}}) . " to go)" );
-            }
-        }
-    }
-}
-
-# ---------------------------------------------------------------------------------------------
-#
-# read_pipe_
-#
-# reads a single message from a pipe in a cross-platform way.
-# returns undef if the pipe has no message
-#
-# $handle   The handle of the pipe to read
-#
-# ---------------------------------------------------------------------------------------------
-sub read_pipe_
-{
-    my ($self, $handle) = @_;
-
-    if ( $^O eq "MSWin32" ) {
-
-        # bypasses bug in -s $pipe under ActivePerl
-
-        my $message;         # PROFILE PLATFORM START MSWin32
-
-        if ( &{ $self->{pipeready_} }($handle) ) {
-
-            # add data to the pipe cache whenever the pipe is ready
-
-            sysread($handle, my $string, -s $handle);
-
-            # push messages onto the end of our cache
-
-            $self->{pipe_cache__} .= $string;
-        }
-
-        # pop the oldest message;
-
-        $message = $1 if ($self->{pipe_cache__} =~ s/(.*?\n)//);
-
-        return $message;        # PROFILE PLATFORM STOP
-    } else {
-
-        # do things normally
-
-        if ( &{ $self->{pipeready_} }($handle) ) {
-            return <$handle>;
-        }
-    }
-
-    return undef;
-}
-
-# ---------------------------------------------------------------------------------------------
-#
-# flush_child_data_
-#
-# Called to flush data from the pipe of each child as we go, I did this because there
-# appears to be a problem on Windows where the pipe gets a lot of read data in it and
-# then causes the child not to be terminated even though we are done.  Also this is nice
-# because we deal with the statistics as we go
-#
-# $handle   The handle of the child's pipe
-#
-# ---------------------------------------------------------------------------------------------
-sub flush_child_data_
-{
-    my ( $self, $handle ) = @_;
-
-    my $stats_changed = 0;
-    my $message;
-
-    while ( ($message = $self->read_pipe_( $handle )) && defined($message) )
-    {
-        $message =~ s/[\r\n]//g;
-
-        $self->log_( "Child proxy message $message" );
-
-        if ( $message =~ /CLASS:([^ ]*) ([^ ]*)/ ) {
-            $self->{classifier__}->classified( $2, $1 );
-        }
-
-        if ( $message =~ /NEWFL:(.*)/ ) {
-            $self->mq_post_( 'NEWFL', $1, '' );
-        }
-
-        if ( $message =~ /LOGIN:(.*)/ ) {
-            $self->mq_post_( 'LOGIN', $1, '' );
-        }
-    }
 }
 
 # ---------------------------------------------------------------------------------------------
@@ -320,66 +204,58 @@ sub service
 {
     my ( $self ) = @_;
 
-    # See if any of the children have passed up statistics data through their
-    # pipes and deal with it now
+    # Accept a connection from a client trying to use us as the mail
+    # server.  We service one client at a time and all others get
+    # queued up to be dealt with later.  We check the alive boolean
+    # here to make sure we are still allowed to operate. See if
+    # there's a connection waiting on the $server by getting the list
+    # of handles with data to read, if the handle is the server then
+    # we're off.
 
-    for my $kid (keys %{$self->{children__}}) {
-        $self->{flush_child_data_}( $self, $self->{children__}{$kid} );
-    }
-
-    # Accept a connection from a client trying to use us as the mail server.  We service one client at a time
-    # and all others get queued up to be dealt with later.  We check the alive boolean here to make sure we
-    # are still allowed to operate. See if there's a connection waiting on the $server by getting the list of
-    # handles with data to read, if the handle is the server then we're off.
-
-    if ( ( defined( $self->{selector__}->can_read(0) ) ) && ( $self->{alive_} ) ) {
+    if ( ( defined( $self->{selector__}->can_read(0) ) ) &&
+         ( $self->{alive_} ) ) {
         if ( my $client = $self->{server__}->accept() ) {
 
             # Check to see if we have obtained a session key yet
 
             if ( $self->{api_session__} eq '' ) {
-                $self->{api_session__} = $self->{classifier__}->get_session_key( 'admin', '' );
+                $self->{api_session__} =
+                    $self->{classifier__}->get_session_key( 'admin', '' );
 	    }
 
-            # Check that this is a connection from the local machine, if it's not then we drop it immediately
-            # without any further processing.  We don't want to act as a proxy for just anyone's email
+            # Check that this is a connection from the local machine,
+            # if it's not then we drop it immediately without any
+            # further processing.  We don't want to act as a proxy for
+            # just anyone's email
 
-            my ( $remote_port, $remote_host ) = sockaddr_in( $client->peername() );
+            my ( $remote_port, $remote_host ) = sockaddr_in(
+                                                    $client->peername() );
 
-            if  ( ( $self->config_( 'local' ) == 0 ) || ( $remote_host eq inet_aton( "127.0.0.1" ) ) ) {
+            if  ( ( ( $self->config_( 'local' ) || 0 ) == 0 ) ||
+                    ( $remote_host eq inet_aton( "127.0.0.1" ) ) ) {
 
-                # Now that we have a good connection to the client fork a subprocess to handle the communication
-                # and set the socket to binmode so that no CRLF translation goes on
-
-                $self->global_config_( 'download_count', $self->global_config_( 'download_count' ) + 1 );
-
-                # If we have force_fork turned on then we will do a fork, otherwise we will handle this
-                # inline, in the inline case we need to create the two ends of a pipe that will be used
-                # as if there was a child process
+                # If we have force_fork turned on then we will do a
+                # fork, otherwise we will handle this inline, in the
+                # inline case we need to create the two ends of a pipe
+                # that will be used as if there was a child process
 
                 binmode( $client );
 
                 if ( $self->config_( 'force_fork' ) ) {
                     my ( $pid, $pipe ) = &{$self->{forker_}};
 
-                    # If we are in the parent process then push the pipe handle onto the children list
-
-                    if ( ( defined( $pid ) ) && ( $pid != 0 ) ) {
-                        $self->{children__}{$pid} = $pipe;
-                    }
-
-                    # If we fail to fork, or are in the child process then process this request
+                    # If we fail to fork, or are in the child process
+                    # then process this request
 
                     if ( !defined( $pid ) || ( $pid == 0 ) ) {
-                        $self->{child_}( $self, $client, $self->global_config_( 'download_count' ), $pipe, 0, $pid, $self->{api_session__} );
-                        $self->{flush_child_data_}( $self, $pipe );
+                        $self->{child_}( $self, $client,
+                            $self->{api_session__} );
                         exit(0) if ( defined( $pid ) );
                     }
 	        } else {
                     pipe my $reader, my $writer;
 
-                    $self->{child_}( $self, $client, $self->global_config_( 'download_count' ), $writer, $reader, $$, $self->{api_session__} );
-                    $self->{flush_child_data_}( $self, $reader );
+                    $self->{child_}( $self, $client, $self->{api_session__} );
                     close $reader;
                 }
             }
@@ -391,24 +267,6 @@ sub service
     return 1;
 }
 
-
-# ---------------------------------------------------------------------------------------------
-#
-# yield_
-#
-# Called by a proxy child process to allow the parent to do work, this only does anything
-# in the case where we didn't fork for the child process
-#
-# ---------------------------------------------------------------------------------------------
-sub yield_
-{
-    my ( $self, $pipe, $pid ) = @_;
-
-    if ( $pid != 0 ) {
-        $self->{flush_child_data_}( $self, $pipe )
-    }
-}
-
 # ---------------------------------------------------------------------------------------------
 #
 # forked
@@ -424,11 +282,6 @@ sub forked
     my ( $self ) = @_;
 
     close $self->{server__};
-
-    for my $kid (keys %{$self->{children__}}) {
-        close $self->{children__}{$kid};
-        delete $self->{children__}{$kid};
-    }
 }
 
 # ---------------------------------------------------------------------------------------------
@@ -446,8 +299,8 @@ sub tee_
     my ( $self, $socket, $text ) = @_;
 
     # Send the message to the debug output and then send it to the appropriate socket
-    $self->log_( $text );
-    print $socket $text;
+    $self->log_( 1, $text );
+    print $socket $text; # don't print if $socket undef
 }
 
 # ---------------------------------------------------------------------------------------------
@@ -477,7 +330,7 @@ sub echo_to_regexp_
                 $self->tee_( $client, $line );
             }
         } else {
-            $self->log_("Suppressed: $line");
+            $self->log_( 2, "Suppressed: $line" );
         }
 
 	if ( $line =~ $regexp ) {
@@ -614,6 +467,7 @@ sub echo_response_
 # $client      The handle to the mail client
 # $hostname    The host name of the remote server
 # $port        The port
+# $ssl         If set to 1 then the connection to the remote is established using SSL
 #
 # Check that we are connected to $hostname on port $port putting the open handle in $mail.
 # Any messages need to be sent to $client
@@ -621,27 +475,50 @@ sub echo_response_
 # ---------------------------------------------------------------------------------------------
 sub verify_connected_
 {
-    my ( $self, $mail, $client, $hostname, $port ) = @_;
+    my ( $self, $mail, $client, $hostname, $port, $ssl ) = @_;
+
+    $ssl = 0 if ( !defined( $ssl ) );
 
     # Check to see if we are already connected
     return $mail if ( $mail && $mail->connected );
 
-    # Connect to the real mail server on the standard port
-   $mail = IO::Socket::INET->new( # PROFILE BLOCK START
-                Proto    => "tcp",
-                PeerAddr => $hostname,
-                PeerPort => $port ); # PROFILE BLOCK STOP
+    # Connect to the real mail server on the standard port, if we are using
+    # SOCKS then go through the proxy server
+
+    if ( $self->config_( 'socks_server' ) ne '' ) {
+        require IO::Socket::Socks;
+        $mail = IO::Socket::Socks->new( # PROFILE BLOCK START
+                    ProxyAddr => $self->config_( 'socks_server' ),
+                    ProxyPort => $self->config_( 'socks_port' ),
+                    ConnectAddr  => $hostname,
+                    ConnectPort  => $port ); # PROFILE BLOCK STOP
+    } else {
+        if ( $ssl ) {
+            require IO::Socket::SSL;
+            $mail = IO::Socket::SSL->new( # PROFILE BLOCK START
+                        Proto    => "tcp",
+                        PeerAddr => $hostname,
+                        PeerPort => $port ); # PROFILE BLOCK STOP
+	} else {
+            $mail = IO::Socket::INET->new( # PROFILE BLOCK START
+                        Proto    => "tcp",
+                        PeerAddr => $hostname,
+                        PeerPort => $port ); # PROFILE BLOCK STOP
+        }
+    }
 
     # Check that the connect succeeded for the remote server
     if ( $mail ) {
         if ( $mail->connected )  {
 
-            $self->log_( "Connected to $hostname:$port timeout " . $self->global_config_( 'timeout' ) );
+            $self->log_( 0, "Connected to $hostname:$port timeout " . $self->global_config_( 'timeout' ) );
 
             # Set binmode on the socket so that no translation of CRLF
             # occurs
 
-            binmode( $mail );
+            if ( !$ssl ) {
+                binmode( $mail );
+	    }
 
             # Wait 10 seconds for a response from the remote server and if
             # there isn't one then give up trying to connect
@@ -672,7 +549,9 @@ sub verify_connected_
                 }
             }
 
-            $self->log_( "Connection returned: $buf" );
+            $self->log_( 1, "Connection returned: $buf" );
+
+            $self->{connect_banner__} = $buf;
 
             # Clean up junk following a newline
 
@@ -690,7 +569,60 @@ sub verify_connected_
     return undef;
 }
 
-# SETTER
+# ---------------------------------------------------------------------------------------------
+#
+# configure_item
+#
+#    $name            The name of the item being configured, was passed in by the call
+#                     to register_configuration_item
+#    $templ           The loaded template
+#
+# ---------------------------------------------------------------------------------------------
+sub configure_item
+{
+    my ( $self, $name, $templ ) = @_;
+
+    $templ->param( 'Socks_Widget_Name' => $self->name() );
+    $templ->param( 'Socks_Server'      => $self->config_( 'socks_server' ) );
+    $templ->param( 'Socks_Port'        => $self->config_( 'socks_port'   ) );
+}
+
+# ---------------------------------------------------------------------------------------------
+#
+# validate_item
+#
+#    $name            The name of the item being configured, was passed in by the call
+#                     to register_configuration_item
+#    $templ           The loaded template
+#    $language        Reference to the hash holding the current language
+#    $form            Hash containing all form items
+#
+#  Must return the HTML for this item
+# ---------------------------------------------------------------------------------------------
+sub validate_item
+{
+    my ( $self, $name, $templ, $language, $form ) = @_;
+
+    my $me = $self->name();
+
+    if ( defined($$form{"$me" . "_socks_port"}) ) {
+        if ( ( $$form{"$me" . "_socks_port"} >= 1 ) && ( $$form{"$me" . "_socks_port"} < 65536 ) ) {
+            $self->config_( 'socks_port', $$form{"$me" . "_socks_port"} );
+            $templ->param( 'Socks_Widget_If_Port_Updated' => 1 );
+            $templ->param( 'Socks_Widget_Port_Updated' => sprintf( $$language{Configuration_SOCKSPortUpdate}, $self->config_( 'socks_port' ) ) );
+        } else {
+            $templ->param( 'Socks_Widget_If_Port_Error' => 1 );
+        }
+    }
+
+    if ( defined($$form{"$me" . "_socks_server"}) ) {
+        $self->config_( 'socks_server', $$form{"$me" . "_socks_server"} );
+        $templ->param( 'Socks_Widget_If_Server_Updated' => 1 );
+        $templ->param( 'Socks_Widget_Server_Updated' => sprintf( $$language{Configuration_SOCKSServerUpdate}, $self->config_( 'socks_server' ) ) );
+    }
+}
+
+# SETTERS
 
 sub classifier
 {
@@ -699,4 +631,11 @@ sub classifier
     $self->{classifier__} = $classifier;
 }
 
+sub history
+{
+    my ( $self, $history ) = @_;
+
+    $self->{history__} = $history;
+}
+
 1;
