@@ -3,7 +3,7 @@
 # This package contains an HTTP server used as a base class for other
 # modules that service requests over HTTP (e.g. the UI)
 #
-# Copyright (c) 2001-2003 John Graham-Cumming
+# Copyright (c) 2001-2004 John Graham-Cumming
 #
 #   This file is part of POPFile
 #
@@ -53,13 +53,13 @@ sub new
     return $self;
 }
 
-# ---------------------------------------------------------------------------------------------
+# ---------------------------------------------------------------------------
 #
 # start
 #
 # Called to start the HTTP interface running
 #
-# ---------------------------------------------------------------------------------------------
+# ---------------------------------------------------------------------------
 sub start
 {
     my ( $self ) = @_;
@@ -119,61 +119,75 @@ sub service
 
     my $code = 1;
 
-    # See if there's a connection waiting for us, if there is we accept it handle a single
-    # request and then exit
+    # See if there's a connection waiting for us, if there is we
+    # accept it handle a single request and then exit
+
     my ( $ready ) = $self->{selector_}->can_read(0);
 
     # Handle HTTP requests for the UI
+
     if ( ( defined( $ready ) ) && ( $ready == $self->{server_} ) ) {
 
         if ( my $client = $self->{server_}->accept() ) {
 
-            # Check that this is a connection from the local machine, if it's not then we drop it immediately
-            # without any further processing.  We don't want to allow remote users to admin POPFile
+            # Check that this is a connection from the local machine,
+            # if it's not then we drop it immediately without any
+            # further processing.  We don't want to allow remote users
+            # to admin POPFile
 
             my ( $remote_port, $remote_host ) = sockaddr_in( $client->peername() );
 
             if ( ( $self->config_( 'local' ) == 0 ) ||                # PROFILE BLOCK START
                  ( $remote_host eq inet_aton( "127.0.0.1" ) ) ) {     # PROFILE BLOCK STOP
 
-                # Read the request line (GET or POST) from the client and if we manage to do that
-                # then read the rest of the HTTP headers grabbing the Content-Length and using
-                # it to read any form POST content into $content
+                # Read the request line (GET or POST) from the client
+                # and if we manage to do that then read the rest of
+                # the HTTP headers grabbing the Content-Length and
+                # using it to read any form POST content into $content
 
                 $client->autoflush(1);
 
-                if ( ( defined( $client ) ) && ( my $request = $self->slurp_( $client ) ) ) {
+                if ( ( defined( $client ) ) &&
+                     ( my $request = $self->slurp_( $client ) ) ) {
                     my $content_length = 0;
                     my $content;
 
+                    $self->log_( 2, $request );
+
                     while ( my $line = $self->slurp_( $client ) )  {
                         $content_length = $1 if ( $line =~ /Content-Length: (\d+)/i );
 
-                        # Discovered that Norton Internet Security was adding
-                        # HTTP headers of the form
+                        # Discovered that Norton Internet Security was
+                        # adding HTTP headers of the form
                         #
                         # ~~~~~~~~~~~~~~: ~~~~~~~~~~~~~
                         #
-                        # which we were not recognizing as valid (surprise,
-                        # surprise) and this was messing about our handling
-                        # of POST data.  Changed the end of header identification
-                        # to any line that does not contain a :
+                        # which we were not recognizing as valid
+                        # (surprise, surprise) and this was messing
+                        # about our handling of POST data.  Changed
+                        # the end of header identification to any line
+                        # that does not contain a :
 
                         last                 if ( $line !~ /:/ );
                     }
 
                     if ( $content_length > 0 ) {
-                        $content = $self->slurp_buffer_( $client, $content_length );
+                        $content = $self->slurp_buffer_( $client,
+                            $content_length );
+                        $self->log_( 2, $content );
                     }
 
                     if ( $request =~ /^(GET|POST) (.*) HTTP\/1\./i ) {
                         $code = $self->handle_url( $client, $2, $1, $content );
+                        $self->log_( 2,
+                            "HTTP handle_url returned code $code\n" );
                     } else {
-                        http_error_( $self, $client, 500 );
+                        $self->http_error_( $client, 500 );
                     }
                 }
             }
 
+            $self->log_( 2, "Close HTTP connection on $client\n" );
             $self->done_slurp_( $client );
             close $client;
         }
@@ -252,7 +266,8 @@ sub parse_form_
 
         $self->{form_}{$arg} =~ s/%([0-9A-F][0-9A-F])/chr hex $1/gie;
 
-        # Push the value onto an array to allow for multiple values of the same name
+        # Push the value onto an array to allow for multiple values of
+        # the same name
 
         if ( $need_array ) {
             push( @{ $self->{form_}{$arg . "_array"} }, $self->{form_}{$arg} );
@@ -313,6 +328,8 @@ sub http_error_
 {
     my ( $self, $client, $error ) = @_;
 
+    $self->log_( 0, "HTTP error $error returned" );
+
     print $client "HTTP/1.0 $error Error$eol$eol";
 }
 
@@ -333,7 +350,7 @@ sub http_file_
     my ( $self, $client, $file, $type ) = @_;
     my $contents = '';
 
-    if ( open FILE, "<$file" ) {
+    if ( defined( $file ) && ( open FILE, "<$file" ) ) {
 
         binmode FILE;
         while (<FILE>) {
@@ -341,9 +358,10 @@ sub http_file_
         }
         close FILE;
 
-        # To prevent the browser for continuously asking for file handled in this way
-        # we calculate the current date and time plus 1 hour to give the browser
-        # cache 1 hour to keep things like graphics and style sheets in cache.
+        # To prevent the browser for continuously asking for file
+        # handled in this way we calculate the current date and time
+        # plus 1 hour to give the browser cache 1 hour to keep things
+        # like graphics and style sheets in cache.
 
         my @day   = ( 'Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat' );
         my @month = ( 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec' );
@@ -364,4 +382,14 @@ sub http_file_
     }
 }
 
+sub history
+{
+    my ( $self, $value ) = @_;
+
+    if ( defined( $value ) ) {
+        $self->{history__} = $value;
+    }
+
+    return $self->{history__};
+}
 
